# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA

from roboid.runner import Runner
from roboid.util import Util
from roboid.model import Robot


class HamsterS(Robot):
    ID = 'kr.robomation.physical.hamster.s'

    LEFT_WHEEL = 0x00400000
    RIGHT_WHEEL = 0x00400001
    BUZZER = 0x00400002
    OUTPUT_A = 0x00400003
    OUTPUT_B = 0x00400004
    NOTE = 0x00400008
    LINE_TRACER_MODE = 0x00400009
    LINE_TRACER_SPEED = 0x0040000a
    IO_MODE_A = 0x0040000b
    IO_MODE_B = 0x0040000c
    CONFIG_PROXIMITY = 0x0040000d
    CONFIG_GRAVITY = 0x0040000e
    CONFIG_BAND_WIDTH = 0x0040000f

    LEFT_RGB = 0x00E00100
    RIGHT_RGB = 0x00E00101
    PULSE = 0x00E00102
    SOUND = 0x00E00103
    LINE_TRACER_GAIN = 0x00E00104
    WRITE_SERIAL = 0x00E00105
    MOTION = 0x00E00106
    MOTOR_MODE = 0x00E00107

    SIGNAL_STRENGTH = 0x00400010
    LEFT_PROXIMITY = 0x00400011
    RIGHT_PROXIMITY = 0x00400012
    LEFT_FLOOR = 0x00400013
    RIGHT_FLOOR = 0x00400014
    ACCELERATION = 0x00400015
    LIGHT = 0x00400016
    TEMPERATURE = 0x00400017
    INPUT_A = 0x00400018
    INPUT_B = 0x00400019
    LINE_TRACER_STATE = 0x0040001a
    TILT = 0x0040001b
    BATTERY_STATE = 0x0040001c

    FREE_FALL = 0x00E00200
    TAP = 0x00E00201
    READ_SERIAL = 0x00E00202
    PULSE_COUNT = 0x00E00203
    WHEEL_STATE = 0x00E00204
    SOUND_STATE = 0x00E00205
    SERIAL_STATE = 0x00E00206

    LINE_TRACER_MODE_OFF = 0
    LINE_TRACER_MODE_BLACK_LEFT_SENSOR = 1
    LINE_TRACER_MODE_BLACK_RIGHT_SENSOR = 2
    LINE_TRACER_MODE_BLACK_BOTH_SENSORS = 3
    LINE_TRACER_MODE_BLACK_TURN_LEFT = 4
    LINE_TRACER_MODE_BLACK_TURN_RIGHT = 5
    LINE_TRACER_MODE_BLACK_MOVE_FORWARD = 6
    LINE_TRACER_MODE_BLACK_UTURN = 7
    LINE_TRACER_MODE_WHITE_LEFT_SENSOR = 8
    LINE_TRACER_MODE_WHITE_RIGHT_SENSOR = 9
    LINE_TRACER_MODE_WHITE_BOTH_SENSORS = 10
    LINE_TRACER_MODE_WHITE_TURN_LEFT = 11
    LINE_TRACER_MODE_WHITE_TURN_RIGHT = 12
    LINE_TRACER_MODE_WHITE_MOVE_FORWARD = 13
    LINE_TRACER_MODE_WHITE_UTURN = 14

    IO_MODE_ANALOG_INPUT = 0
    IO_MODE_ANALOG_INPUT_RELATIVE = 0
    IO_MODE_DIGITAL_INPUT = 1
    IO_MODE_DIGITAL_INPUT_TRI_STATE = 1
    IO_MODE_DIGITAL_INPUT_PULL_UP = 2
    IO_MODE_DIGITAL_INPUT_PULL_DOWN = 3
    IO_MODE_ANALOG_INPUT_ABSOLUTE = 4
    IO_MODE_SERVO_OUTPUT = 8
    IO_MODE_PWM_OUTPUT = 9
    IO_MODE_DIGITAL_OUTPUT = 10
    IO_MODE_SERIAL_9600 = 176
    IO_MODE_SERIAL_14400 = 177
    IO_MODE_SERIAL_19200 = 178
    IO_MODE_SERIAL_28800 = 179
    IO_MODE_SERIAL_38400 = 180
    IO_MODE_SERIAL_57600 = 181
    IO_MODE_SERIAL_76800 = 182
    IO_MODE_SERIAL_115200 = 183

    NOTE_OFF = 0
    NOTE_A_0 = 1
    NOTE_A_SHARP_0 = 2
    NOTE_B_FLAT_0 = 2
    NOTE_B_0 = 3
    NOTE_C_1 = 4
    NOTE_C_SHARP_1 = 5
    NOTE_D_FLAT_1 = 5
    NOTE_D_1 = 6
    NOTE_D_SHARP_1 = 7
    NOTE_E_FLAT_1 = 7
    NOTE_E_1 = 8
    NOTE_F_1 = 9
    NOTE_F_SHARP_1 = 10
    NOTE_G_FLAT_1 = 10
    NOTE_G_1 = 11
    NOTE_G_SHARP_1 = 12
    NOTE_A_FLAT_1 = 12
    NOTE_A_1 = 13
    NOTE_A_SHARP_1 = 14
    NOTE_B_FLAT_1 = 14
    NOTE_B_1 = 15
    NOTE_C_2 = 16
    NOTE_C_SHARP_2 = 17
    NOTE_D_FLAT_2 = 17
    NOTE_D_2 = 18
    NOTE_D_SHARP_2 = 19
    NOTE_E_FLAT_2 = 19
    NOTE_E_2 = 20
    NOTE_F_2 = 21
    NOTE_F_SHARP_2 = 22
    NOTE_G_FLAT_2 = 22
    NOTE_G_2 = 23
    NOTE_G_SHARP_2 = 24
    NOTE_A_FLAT_2 = 24
    NOTE_A_2 = 25
    NOTE_A_SHARP_2 = 26
    NOTE_B_FLAT_2 = 26
    NOTE_B_2 = 27
    NOTE_C_3 = 28
    NOTE_C_SHARP_3 = 29
    NOTE_D_FLAT_3 = 29
    NOTE_D_3 = 30
    NOTE_D_SHARP_3 = 31
    NOTE_E_FLAT_3 = 31
    NOTE_E_3 = 32
    NOTE_F_3 = 33
    NOTE_F_SHARP_3 = 34
    NOTE_G_FLAT_3 = 34
    NOTE_G_3 = 35
    NOTE_G_SHARP_3 = 36
    NOTE_A_FLAT_3 = 36
    NOTE_A_3 = 37
    NOTE_A_SHARP_3 = 38
    NOTE_B_FLAT_3 = 38
    NOTE_B_3 = 39
    NOTE_C_4 = 40
    NOTE_C_SHARP_4 = 41
    NOTE_D_FLAT_4 = 41
    NOTE_D_4 = 42
    NOTE_D_SHARP_4 = 43
    NOTE_E_FLAT_4 = 43
    NOTE_E_4 = 44
    NOTE_F_4 = 45
    NOTE_F_SHARP_4 = 46
    NOTE_G_FLAT_4 = 46
    NOTE_G_4 = 47
    NOTE_G_SHARP_4 = 48
    NOTE_A_FLAT_4 = 48
    NOTE_A_4 = 49
    NOTE_A_SHARP_4 = 50
    NOTE_B_FLAT_4 = 50
    NOTE_B_4 = 51
    NOTE_C_5 = 52
    NOTE_C_SHARP_5 = 53
    NOTE_D_FLAT_5 = 53
    NOTE_D_5 = 54
    NOTE_D_SHARP_5 = 55
    NOTE_E_FLAT_5 = 55
    NOTE_E_5 = 56
    NOTE_F_5 = 57
    NOTE_F_SHARP_5 = 58
    NOTE_G_FLAT_5 = 58
    NOTE_G_5 = 59
    NOTE_G_SHARP_5 = 60
    NOTE_A_FLAT_5 = 60
    NOTE_A_5 = 61
    NOTE_A_SHARP_5 = 62
    NOTE_B_FLAT_5 = 62
    NOTE_B_5 = 63
    NOTE_C_6 = 64
    NOTE_C_SHARP_6 = 65
    NOTE_D_FLAT_6 = 65
    NOTE_D_6 = 66
    NOTE_D_SHARP_6 = 67
    NOTE_E_FLAT_6 = 67
    NOTE_E_6 = 68
    NOTE_F_6 = 69
    NOTE_F_SHARP_6 = 70
    NOTE_G_FLAT_6 = 70
    NOTE_G_6 = 71
    NOTE_G_SHARP_6 = 72
    NOTE_A_FLAT_6 = 72
    NOTE_A_6 = 73
    NOTE_A_SHARP_6 = 74
    NOTE_B_FLAT_6 = 74
    NOTE_B_6 = 75
    NOTE_C_7 = 76
    NOTE_C_SHARP_7 = 77
    NOTE_D_FLAT_7 = 77
    NOTE_D_7 = 78
    NOTE_D_SHARP_7 = 79
    NOTE_E_FLAT_7 = 79
    NOTE_E_7 = 80
    NOTE_F_7 = 81
    NOTE_F_SHARP_7 = 82
    NOTE_G_FLAT_7 = 82
    NOTE_G_7 = 83
    NOTE_G_SHARP_7 = 84
    NOTE_A_FLAT_7 = 84
    NOTE_A_7 = 85
    NOTE_A_SHARP_7 = 86
    NOTE_B_FLAT_7 = 86
    NOTE_B_7 = 87
    NOTE_C_8 = 88
    
    SOUND_OFF = 0
    SOUND_BEEP = 1
    SOUND_RANDOM_BEEP = 2
    SOUND_NOISE = 10
    SOUND_SIREN = 3
    SOUND_ENGINE = 4
    SOUND_CHOP = 11
    SOUND_ROBOT = 5
    SOUND_DIBIDIBIDIP = 8
    SOUND_GOOD_JOB = 9
    SOUND_HAPPY = 12
    SOUND_ANGRY = 13
    SOUND_SAD = 14
    SOUND_SLEEP = 15
    SOUND_MARCH = 6
    SOUND_BIRTHDAY = 7

    BATTERY_NORMAL = 2
    BATTERY_LOW = 1
    BATTERY_EMPTY = 0

    MOTOR_AUTO = 0
    MOTOR_HALF = 1
    MOTOR_WAVE = 2
    MOTOR_FULL = 3
    
    _RGBS = {
        'red': (255, 0, 0),
        'orange': (255, 63, 0),
        'yellow': (255, 255, 0),
        'green': (0, 255, 0),
        'sky blue': (0, 255, 255),
        'skyblue': (0, 255, 255),
        'cyan': (0, 255, 255),
        'blue': (0, 0, 255),
        'violet': (63, 0, 255),
        'purple': (255, 0, 255),
        'magenta': (255, 0, 255),
        'white': (255, 255, 255),
        0: (0, 0, 0),
        1: (0, 0, 255),
        2: (0, 255, 0),
        3: (0, 255, 255),
        4: (255, 0, 0),
        5: (255, 0, 255),
        6: (255, 255, 0),
        7: (255, 255, 255)
    }

    _index = 0

    def __init__(self, port_name=None):
        super(HamsterS, self).__init__(HamsterS.ID, 'HamsterS', HamsterS._index)
        HamsterS._index += 1
        self._bpm = 60
        self._init(port_name)

    def dispose(self):
        self._roboid._dispose()
        Runner.unregister_robot(self)

    def reset(self):
        self._roboid._reset()

    def _init(self, port_name):
        from roboid.hamster_s_roboid import HamsterSRoboid
        self._roboid = HamsterSRoboid(self.get_index())
        self._add_roboid(self._roboid)
        Runner.register_robot(self)
        Runner.start()
        self._roboid._init(port_name)

    def find_device_by_id(self, device_id):
        return self._roboid.find_device_by_id(device_id)

    def _request_motoring_data(self):
        self._roboid._request_motoring_data()

    def _update_sensory_device_state(self):
        self._roboid._update_sensory_device_state()

    def _update_motoring_device_state(self):
        self._roboid._update_motoring_device_state()

    def _notify_sensory_device_data_changed(self):
        self._roboid._notify_sensory_device_data_changed()

    def _notify_motoring_device_data_changed(self):
        self._roboid._notify_motoring_device_data_changed()

    def cm_to_pulse(self, cm):
        if isinstance(cm, (int, float)):
            return Util.round(cm * 136.15)
        else:
            return 0

    def calc_turn_pulse(self, degree):
        if isinstance(degree, (int, float)):
            return Util.round(degree * 3.975)
        else:
            return 0

    def calc_pivot_pulse(self, degree):
        if isinstance(degree, (int, float)):
            return Util.round(degree * 3.975 * 2)
        else:
            return 0

    def calc_inner_velocity(self, outer_velocity, radius):
        if isinstance(outer_velocity, (int, float)) and isinstance(radius, (int, float)):
            return outer_velocity * (radius - 1.6795) / (radius + 1.6795)
        else:
            return 0

    def calc_swing_pulse(self, degree, radius):
        if isinstance(degree, (int, float)) and isinstance(radius, (int, float)):
            return Util.round(degree * 3.82 * (radius + 1.6795) / 1.6795)
        else:
            return 0

    def wheels(self, left_velocity, right_velocity=None):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(left_velocity, (int, float)):
            self.write(HamsterS.LEFT_WHEEL, left_velocity)
        if isinstance(right_velocity, (int, float)):
            self.write(HamsterS.RIGHT_WHEEL, right_velocity)
        else:
            if isinstance(left_velocity, (int, float)):
                self.write(HamsterS.RIGHT_WHEEL, left_velocity)

    def left_wheel(self, velocity):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(velocity, (int, float)):
            self.write(HamsterS.LEFT_WHEEL, velocity)

    def right_wheel(self, velocity):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(velocity, (int, float)):
            self.write(HamsterS.RIGHT_WHEEL, velocity)

    def stop(self):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        self.write(HamsterS.LEFT_WHEEL, 0)
        self.write(HamsterS.RIGHT_WHEEL, 0)

    def _evaluate_wheel_state(self):
        return self.e(HamsterS.WHEEL_STATE) and self.read(HamsterS.WHEEL_STATE) == 2

    def move_forward(self, cm=5, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(cm, (int, float)) and isinstance(velocity, (int, float)):
            if cm < 0:
                cm = -cm
                velocity = -velocity
            pulse = self.cm_to_pulse(cm)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def move_forward_sec(self, sec, velocity=30):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            if sec > 0:
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                Runner.wait(sec * 1000)
            self.write(HamsterS.LEFT_WHEEL, 0)
            self.write(HamsterS.RIGHT_WHEEL, 0)

    def move_forward_pulse(self, pulse, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            pulse = Util.round(pulse)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def move_backward(self, cm=5, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(cm, (int, float)) and isinstance(velocity, (int, float)):
            if cm < 0:
                cm = -cm
                velocity = -velocity
            pulse = self.cm_to_pulse(cm)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, -velocity)
                self.write(HamsterS.RIGHT_WHEEL, -velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def move_backward_sec(self, sec, velocity=30):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            if sec > 0:
                self.write(HamsterS.LEFT_WHEEL, -velocity)
                self.write(HamsterS.RIGHT_WHEEL, -velocity)
                Runner.wait(sec * 1000)
            self.write(HamsterS.LEFT_WHEEL, 0)
            self.write(HamsterS.RIGHT_WHEEL, 0)

    def move_backward_pulse(self, pulse, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            pulse = Util.round(pulse)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, -velocity)
                self.write(HamsterS.RIGHT_WHEEL, -velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def turn_left(self, degree=90, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(degree, (int, float)) and isinstance(velocity, (int, float)):
            if degree < 0:
                degree = -degree;
                velocity = -velocity;
            pulse = self.calc_turn_pulse(degree)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, -velocity)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def turn_left_sec(self, sec, velocity=30):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            if sec > 0:
                self.write(HamsterS.LEFT_WHEEL, -velocity)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                Runner.wait(sec * 1000)
            self.write(HamsterS.LEFT_WHEEL, 0)
            self.write(HamsterS.RIGHT_WHEEL, 0)

    def turn_left_pulse(self, pulse, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            pulse = Util.round(pulse)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, -velocity)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def turn_right(self, degree=90, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(degree, (int, float)) and isinstance(velocity, (int, float)):
            if degree < 0:
                degree = -degree;
                velocity = -velocity;
            pulse = self.calc_turn_pulse(degree)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, -velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def turn_right_sec(self, sec, velocity=30):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            if sec > 0:
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, -velocity)
                Runner.wait(sec * 1000)
            self.write(HamsterS.LEFT_WHEEL, 0)
            self.write(HamsterS.RIGHT_WHEEL, 0)

    def turn_right_pulse(self, pulse, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            pulse = Util.round(pulse)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, -velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def pivot_left_wheel(self, degree, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(degree, (int, float)) and isinstance(velocity, (int, float)):
            if degree < 0:
                degree = -degree;
                velocity = -velocity;
            pulse = self.calc_pivot_pulse(degree)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def pivot_left_wheel_sec(self, sec, velocity=30):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            if sec > 0:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                Runner.wait(sec * 1000)
            self.write(HamsterS.LEFT_WHEEL, 0)
            self.write(HamsterS.RIGHT_WHEEL, 0)

    def pivot_left_wheel_pulse(self, pulse, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            pulse = Util.round(pulse)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def pivot_right_wheel(self, degree, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(degree, (int, float)) and isinstance(velocity, (int, float)):
            if degree < 0:
                degree = -degree;
                velocity = -velocity;
            pulse = self.calc_pivot_pulse(degree)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def pivot_right_wheel_sec(self, sec, velocity=30):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(sec, (int, float)) and isinstance(velocity, (int, float)):
            if sec > 0:
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                Runner.wait(sec * 1000)
            self.write(HamsterS.LEFT_WHEEL, 0)
            self.write(HamsterS.RIGHT_WHEEL, 0)

    def pivot_right_wheel_pulse(self, pulse, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(pulse, (int, float)) and isinstance(velocity, (int, float)):
            pulse = Util.round(pulse)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def pivot_left_pen(self, degree, velocity=30):
        self.swing_left(degree, 2.51, velocity)

    def pivot_left_pen_sec(self, sec, velocity=30):
        self.swing_left_sec(sec, 2.51, velocity)

    def pivot_left_pen_pulse(self, pulse, velocity=30):
        self.swing_left_pulse(pulse, 2.51, velocity)

    def pivot_right_pen(self, degree, velocity=30):
        self.swing_right(degree, 2.51, velocity)

    def pivot_right_pen_sec(self, sec, velocity=30):
        self.swing_right_sec(sec, 2.51, velocity)

    def pivot_right_pen_pulse(self, pulse, velocity=30):
        self.swing_right_pulse(pulse, 2.51, velocity)

    def swing_left(self, degree, radius, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(degree, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            if radius < 0:
                degree = 0
            inner_velocity = self.calc_inner_velocity(velocity, radius)
            if degree < 0:
                degree = -degree
                velocity = -velocity
                inner_velocity = -inner_velocity
            pulse = self.calc_swing_pulse(degree, radius)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, inner_velocity)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def swing_left_sec(self, sec, radius, velocity=30):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(sec, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            if sec > 0 and radius >= 0:
                inner_velocity = self.calc_inner_velocity(velocity, radius)
                self.write(HamsterS.LEFT_WHEEL, inner_velocity)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                Runner.wait(sec * 1000)
            self.write(HamsterS.LEFT_WHEEL, 0)
            self.write(HamsterS.RIGHT_WHEEL, 0)

    def swing_left_pulse(self, pulse, radius, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(pulse, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            pulse = Util.round(pulse)
            if pulse > 0 and radius >= 0:
                inner_velocity = self.calc_inner_velocity(velocity, radius)
                self.write(HamsterS.LEFT_WHEEL, inner_velocity)
                self.write(HamsterS.RIGHT_WHEEL, velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def swing_right(self, degree, radius, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(degree, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            if radius < 0:
                degree = 0
            inner_velocity = self.calc_inner_velocity(velocity, radius)
            if degree < 0:
                degree = -degree
                velocity = -velocity
                inner_velocity = -inner_velocity
            pulse = self.calc_swing_pulse(degree, radius)
            if pulse > 0:
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, inner_velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def swing_right_sec(self, sec, radius, velocity=30):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(sec, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            if sec > 0 and radius >= 0:
                inner_velocity = self.calc_inner_velocity(velocity, radius)
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, inner_velocity)
                Runner.wait(sec * 1000)
            self.write(HamsterS.LEFT_WHEEL, 0)
            self.write(HamsterS.RIGHT_WHEEL, 0)

    def swing_right_pulse(self, pulse, radius, velocity=30):
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        if isinstance(pulse, (int, float)) and isinstance(radius, (int, float)) and isinstance(velocity, (int, float)):
            pulse = Util.round(pulse)
            if pulse > 0 and radius >= 0:
                inner_velocity = self.calc_inner_velocity(velocity, radius)
                self.write(HamsterS.LEFT_WHEEL, velocity)
                self.write(HamsterS.RIGHT_WHEEL, inner_velocity)
                self.write(HamsterS.PULSE, pulse)
                Runner.wait_until(self._evaluate_wheel_state)
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                self.write(HamsterS.PULSE, 0)

    def _evaluate_line_tracer(self):
        return self.e(HamsterS.LINE_TRACER_STATE) and self.read(HamsterS.LINE_TRACER_STATE) == 0x40

    def line_tracer_mode(self, mode):
        self.write(HamsterS.LEFT_WHEEL, 0)
        self.write(HamsterS.RIGHT_WHEEL, 0)
        self.write(HamsterS.PULSE, 0)
        if isinstance(mode, (int, float)):
            mode = int(mode)
            self.write(HamsterS.LINE_TRACER_MODE, mode)
            if (mode >= 4 and mode <= 7) or (mode >= 11 and mode <= 14):
                Runner.wait_until(self._evaluate_line_tracer)

    def line_tracer_gain(self, gain):
        if isinstance(gain, (int, float)):
            self.write(HamsterS.LINE_TRACER_GAIN, Util.round(gain))

    def line_tracer_speed(self, speed):
        if isinstance(speed, (int, float)):
            self.write(HamsterS.LINE_TRACER_SPEED, Util.round(speed))

    def _evaluate_board_forward(self):
        if self._board_state == 1:
            if self._board_count < 2:
                left_floor = self.read(HamsterS.LEFT_FLOOR)
                right_floor = self.read(HamsterS.RIGHT_FLOOR)
                if left_floor < 50 and right_floor < 50:
                    self._board_count += 1
                else:
                    self._board_count = 0
                diff = left_floor - right_floor
                self.write(HamsterS.LEFT_WHEEL, Util.round(45 + diff * 0.25))
                self.write(HamsterS.RIGHT_WHEEL, Util.round(45 - diff * 0.25))
            else:
                self._board_count = 0
                self._board_state = 2
        elif self._board_state == 2:
            if self._board_count < 10:
                self._board_count += 1
                diff = self.read(HamsterS.LEFT_FLOOR) - self.read(HamsterS.RIGHT_FLOOR)
                self.write(HamsterS.LEFT_WHEEL, Util.round(45 + diff * 0.25))
                self.write(HamsterS.RIGHT_WHEEL, Util.round(45 - diff * 0.25))
            else:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                return True
        return False

    def board_forward(self):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        self.write(HamsterS.LEFT_WHEEL, 45)
        self.write(HamsterS.RIGHT_WHEEL, 45)
        self._board_count = 0
        self._board_state = 1
        Runner.wait_until(self._evaluate_board_forward)

    def _evaluate_board_left(self):
        state = self._board_state
        if state == 1:
            if self._board_count < 2:
                if self.read(HamsterS.LEFT_FLOOR) > 50:
                    self._board_count += 1
            else:
                self._board_count = 0
                self._board_state = 2
        elif state == 2:
            if self.read(HamsterS.LEFT_FLOOR) < 20:
                self._board_state = 3
        elif state == 3:
            if self._board_count < 2:
                if self.read(HamsterS.LEFT_FLOOR) < 20:
                    self._board_count += 1
            else:
                self._board_count = 0
                self._board_state = 4
        elif state == 4:
            if self.read(HamsterS.LEFT_FLOOR) > 50:
                self._board_state = 5
        elif state == 5:
            diff = self.read(HamsterS.LEFT_FLOOR) - self.read(HamsterS.RIGHT_FLOOR)
            if diff > -15:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                return True
            else:
                self.write(HamsterS.LEFT_WHEEL, Util.round(diff * 0.5))
                self.write(HamsterS.RIGHT_WHEEL, -Util.round(diff * 0.5))
        return False

    def board_left(self):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        self.write(HamsterS.LEFT_WHEEL, -45)
        self.write(HamsterS.RIGHT_WHEEL, 45)
        self._board_count = 0
        self._board_state = 1
        Runner.wait_until(self._evaluate_board_left)

    def _evaluate_board_right(self):
        state = self._board_state
        if state == 1:
            if self._board_count < 2:
                if self.read(HamsterS.RIGHT_FLOOR) > 50:
                    self._board_count += 1
            else:
                self._board_count = 0
                self._board_state = 2
        elif state == 2:
            if self.read(HamsterS.RIGHT_FLOOR) < 20:
                self._board_state = 3
        elif state == 3:
            if self._board_count < 2:
                if self.read(HamsterS.RIGHT_FLOOR) < 20:
                    self._board_count += 1
            else:
                self._board_count = 0
                self._board_state = 4
        elif state == 4:
            if self.read(HamsterS.RIGHT_FLOOR) > 50:
                self._board_state = 5
        elif state == 5:
            diff = self.read(HamsterS.RIGHT_FLOOR) - self.read(HamsterS.LEFT_FLOOR)
            if diff > -15:
                self.write(HamsterS.LEFT_WHEEL, 0)
                self.write(HamsterS.RIGHT_WHEEL, 0)
                return True
            else:
                self.write(HamsterS.LEFT_WHEEL, -Util.round(diff * 0.5))
                self.write(HamsterS.RIGHT_WHEEL, Util.round(diff * 0.5))
        return False

    def board_right(self):
        self.write(HamsterS.PULSE, 0)
        self.write(HamsterS.LINE_TRACER_MODE, HamsterS.LINE_TRACER_MODE_OFF)
        self.write(HamsterS.LEFT_WHEEL, 45)
        self.write(HamsterS.RIGHT_WHEEL, -45)
        self._board_count = 0
        self._board_state = 1
        Runner.wait_until(self._evaluate_board_right)

    def rgbs(self, color1, color2=None, color3=None, color4=None, color5=None, color6=None):
        if isinstance(color1, (int, float)):
            color1 = Util.round(color1)
            if isinstance(color2, (int, float)):
                color2 = Util.round(color2)
                if isinstance(color3, (int, float)):
                    color3 = Util.round(color3)
                    if isinstance(color4, (int, float)) and isinstance(color5, (int, float)) and isinstance(color6, (int, float)):
                        color4 = Util.round(color4)
                        color5 = Util.round(color5)
                        color6 = Util.round(color6)
                        self.write(HamsterS.LEFT_RGB, 0, color1)
                        self.write(HamsterS.LEFT_RGB, 1, color2)
                        self.write(HamsterS.LEFT_RGB, 2, color3)
                        self.write(HamsterS.RIGHT_RGB, 0, color4)
                        self.write(HamsterS.RIGHT_RGB, 1, color5)
                        self.write(HamsterS.RIGHT_RGB, 2, color6)
                    else:
                        self.write(HamsterS.LEFT_RGB, 0, color1)
                        self.write(HamsterS.LEFT_RGB, 1, color2)
                        self.write(HamsterS.LEFT_RGB, 2, color3)
                        self.write(HamsterS.RIGHT_RGB, 0, color1)
                        self.write(HamsterS.RIGHT_RGB, 1, color2)
                        self.write(HamsterS.RIGHT_RGB, 2, color3)
                else:
                    self.write(HamsterS.LEFT_RGB, 0, color1)
                    self.write(HamsterS.LEFT_RGB, 1, color1)
                    self.write(HamsterS.LEFT_RGB, 2, color1)
                    self.write(HamsterS.RIGHT_RGB, 0, color2)
                    self.write(HamsterS.RIGHT_RGB, 1, color2)
                    self.write(HamsterS.RIGHT_RGB, 2, color2)
            else:
                self.write(HamsterS.LEFT_RGB, 0, color1)
                self.write(HamsterS.LEFT_RGB, 1, color1)
                self.write(HamsterS.LEFT_RGB, 2, color1)
                self.write(HamsterS.RIGHT_RGB, 0, color1)
                self.write(HamsterS.RIGHT_RGB, 1, color1)
                self.write(HamsterS.RIGHT_RGB, 2, color1)

    def left_rgb(self, color1, color2=None, color3=None):
        if isinstance(color1, (int, float)):
            color1 = Util.round(color1)
            if isinstance(color2, (int, float)) and isinstance(color3, (int, float)):
                color2 = Util.round(color2)
                color3 = Util.round(color3)
                self.write(HamsterS.LEFT_RGB, 0, color1)
                self.write(HamsterS.LEFT_RGB, 1, color2)
                self.write(HamsterS.LEFT_RGB, 2, color3)
            else:
                self.write(HamsterS.LEFT_RGB, 0, color1)
                self.write(HamsterS.LEFT_RGB, 1, color1)
                self.write(HamsterS.LEFT_RGB, 2, color1)

    def right_rgb(self, color1, color2=None, color3=None):
        if isinstance(color1, (int, float)):
            color1 = Util.round(color1)
            if isinstance(color2, (int, float)) and isinstance(color3, (int, float)):
                color2 = Util.round(color2)
                color3 = Util.round(color3)
                self.write(HamsterS.RIGHT_RGB, 0, color1)
                self.write(HamsterS.RIGHT_RGB, 1, color2)
                self.write(HamsterS.RIGHT_RGB, 2, color3)
            else:
                self.write(HamsterS.RIGHT_RGB, 0, color1)
                self.write(HamsterS.RIGHT_RGB, 1, color1)
                self.write(HamsterS.RIGHT_RGB, 2, color1)

    def leds(self, color1, color2=None):
        if color1 in _RGBS:
            color1 = _RGBS[color1]
            self.write(HamsterS.LEFT_RGB, color1)
            if color2 is None:
                self.write(HamsterS.RIGHT_RGB, color1)
            elif color2 in _RGBS:
                self.write(HamsterS.RIGHT_RGB, _RGBS[color2])

    def left_led(self, color):
        if color in _RGBS:
            self.write(HamsterS.LEFT_RGB, _RGBS[color])

    def right_led(self, color):
        if color in _RGBS:
            self.write(HamsterS.RIGHT_RGB, _RGBS[color])

    def beep(self):
        self.write(HamsterS.NOTE, HamsterS.NOTE_OFF)
        self.write(HamsterS.SOUND, 0)
        self.write(HamsterS.BUZZER, 440)
        Runner.wait(200)
        self.write(HamsterS.BUZZER, 0)

    def buzzer(self, hz):
        self.write(HamsterS.NOTE, HamsterS.NOTE_OFF)
        self.write(HamsterS.SOUND, 0)
        if isinstance(hz, (int, float)):
            self.write(HamsterS.BUZZER, hz)

    def tempo(self, bpm):
        if isinstance(bpm, (int, float)):
            if bpm > 0:
                self._bpm = bpm

    def note(self, pitch, beats=None):
        self.write(HamsterS.BUZZER, 0)
        self.write(HamsterS.SOUND, 0)
        if isinstance(pitch, (int, float)):
            pitch = int(pitch)
            if isinstance(beats, (int, float)):
                bpm = self._bpm
                if beats > 0 and bpm > 0:
                    if pitch == 0:
                        self.write(HamsterS.NOTE, HamsterS.NOTE_OFF)
                        Runner.wait(beats * 60 * 1000.0 / bpm)
                    elif pitch > 0:
                        timeout = beats * 60 * 1000.0 / bpm
                        tail = 0
                        if timeout > 100:
                            tail = 100
                        self.write(HamsterS.NOTE, pitch)
                        Runner.wait(timeout - tail)
                        self.write(HamsterS.NOTE, HamsterS.NOTE_OFF)
                        if tail > 0:
                            Runner.wait(tail)
                else:
                    self.write(HamsterS.NOTE, HamsterS.NOTE_OFF)
            elif pitch >= 0:
                self.write(HamsterS.NOTE, pitch)

    def _evaluate_sound(self):
        return self.e(HamsterS.SOUND_STATE) and self.read(HamsterS.SOUND_STATE) == 0

    def sound(self, sound):
        self.write(HamsterS.BUZZER, 0)
        self.write(HamsterS.NOTE, HamsterS.NOTE_OFF)
        if isinstance(sound, (int, float)):
            self.write(HamsterS.SOUND, int(sound))

    def sound_until_done(self, sound, count=1):
        self.write(HamsterS.BUZZER, 0)
        self.write(HamsterS.NOTE, HamsterS.NOTE_OFF)
        if isinstance(sound, (int, float)) and isinstance(count, (int, float)):
            sound = int(sound)
            count = int(count)
            if sound > 0:
                if count > 0:
                    for i in range(count):
                        self.write(HamsterS.SOUND, sound)
                        Runner.wait_until(self._evaluate_sound)
                elif count < 0:
                    while True:
                        self.write(HamsterS.SOUND, sound)
                        Runner.wait_until(self._evaluate_sound)
            else:
                self.write(HamsterS.SOUND, 0)

    def io_mode_a(self, mode):
        if isinstance(mode, (int, float)):
            self.write(HamsterS.IO_MODE_A, int(mode))

    def io_mode_b(self, mode):
        if isinstance(mode, (int, float)):
            self.write(HamsterS.IO_MODE_B, int(mode))

    def output_a(self, value):
        if isinstance(value, (int, float)):
            self.write(HamsterS.OUTPUT_A, Util.round(value))

    def output_b(self, value):
        if isinstance(value, (int, float)):
            self.write(HamsterS.OUTPUT_B, Util.round(value))

    def open_gripper(self):
        self.write(HamsterS.IO_MODE_A, HamsterS.IO_MODE_DIGITAL_OUTPUT)
        self.write(HamsterS.IO_MODE_B, HamsterS.IO_MODE_DIGITAL_OUTPUT)
        self.write(HamsterS.OUTPUT_A, 1)
        self.write(HamsterS.OUTPUT_B, 0)
        Runner.wait(500)

    def close_gripper(self):
        self.write(HamsterS.IO_MODE_A, HamsterS.IO_MODE_DIGITAL_OUTPUT)
        self.write(HamsterS.IO_MODE_B, HamsterS.IO_MODE_DIGITAL_OUTPUT)
        self.write(HamsterS.OUTPUT_A, 0)
        self.write(HamsterS.OUTPUT_B, 1)
        Runner.wait(500)

    def release_gripper(self):
        self.write(HamsterS.IO_MODE_A, HamsterS.IO_MODE_DIGITAL_OUTPUT)
        self.write(HamsterS.IO_MODE_B, HamsterS.IO_MODE_DIGITAL_OUTPUT)
        self.write(HamsterS.OUTPUT_A, 0)
        self.write(HamsterS.OUTPUT_B, 0)

    def write_serial(self, data):
        self._roboid._write_serial(data)

    def read_serial(self, delimiter=None):
        return self._roboid._read_serial(delimiter)

    def serial_rate(self, baud):
        self._roboid._serial_rate(baud)

    def signal_strength(self):
        return self.read(HamsterS.SIGNAL_STRENGTH)

    def left_proximity(self):
        return self.read(HamsterS.LEFT_PROXIMITY)

    def right_proximity(self):
        return self.read(HamsterS.RIGHT_PROXIMITY)

    def left_floor(self):
        return self.read(HamsterS.LEFT_FLOOR)

    def right_floor(self):
        return self.read(HamsterS.RIGHT_FLOOR)

    def acceleration_x(self):
        return self.read(HamsterS.ACCELERATION, 0)

    def acceleration_y(self):
        return self.read(HamsterS.ACCELERATION, 1)

    def acceleration_z(self):
        return self.read(HamsterS.ACCELERATION, 2)

    def light(self):
        return self.read(HamsterS.LIGHT)

    def temperature(self):
        return self.read(HamsterS.TEMPERATURE)

    def input_a(self):
        return self.read(HamsterS.INPUT_A)

    def input_b(self):
        return self.read(HamsterS.INPUT_B)

    def free_fall(self):
        return self.e(HamsterS.FREE_FALL)

    def tap(self):
        return self.e(HamsterS.TAP)

    def tilt(self):
        return self.read(HamsterS.TILT)

    def pulse_count(self):
        return self.read(HamsterS.PULSE_COUNT)

    def battery_state(self):
        return self.read(HamsterS.BATTERY_STATE)